require 'rubygems'
require 'tempfile'
require 'optparse'
require 'crypt/blowfish'
require 'delegate'
require 'yaml'
$:.unshift File.dirname(__FILE__)

class SimplePass < DelegateClass(Hash)
  VERSION = "1.0.1"

  def self.process_args(argv)
    @options = options = {
      :editor => ENV['EDITOR'] || 'nano'
    }
    opts = OptionParser.new do |opt|
      opt.program_name = File.basename $0
      opt.version = SimplePass::VERSION
      opt.summary_indent = ' ' * 4
      opt.banner = <<-EOT
Usage: #{opt.program_name} [options] [domain name]

Where domain name is typically a website domain, e.g. yahoo.com. The domain can
also be any other string, e.g. "Gmail account".

For example, you enter 

    #{opt.program_name} gmail.com
    
If your domain name has spaces, remember to quote it like so:

    #{opt.program_name} 'Gmail account'

for the first time, simplepass will launch your text editor (whatever your
EDITOR environment variable has been set to) and present you with a simple form
that you can fill out to save a login, password, and arbitrary notes for that
domain: 

    gmail.com
    login:
    password:
    notes:

You can leave any of the fields blank. So if you want, you can just fill out 
the 'notes' portion for items that are not web logins, such as credit card
numbers and such.

Once you fill out the fields save the file and exit your editor, simplepass will
parse the information, encrypt it, store it in the database, and delete the
temporary file.

The next time you give the command,
    
    #{opt.program_name} gmail.com

simplepass will display something like this, via the 'less' command: (This is so 

    gmail.com
    login: funnyface
    password: audreyhepburn
    notes: I love gmail! blah... blah...
  
(We pipe the output to 'less' so as to leave no trace of your password in your
console, where someone can find it by scrolling up.)

The first time you launch simplepass, it will ask you to set a master password.
This password will unlock the simplepass database. This database is stored in
a single file: a partly encrypted text file called simplepass.db. This file will be 
saved in the directory in which you invoked the #{opt.program_name} command.

The top part of this file is a message in plain text and just serves to remind
you of the file's purpose. The bottom part is your password database, which is
nothing more than a YAML string encrypted with the Blowfish encryption
algorithm. Your master password is the the key that decrypts this portion of the
simplepass.db file. Do not edit this file directly; any edits may render the
data un-decryptable.
      EOT
      opt.separator nil
      opt.separator "Options:"
      opt.separator nil

      opt.on("--edit", "-e", 
             "Edit the login information for the domain.",
             "This invokes your text editor.") do |value|
        options[:edit] = true
      end
      opt.separator nil
      opt.on("--remove", "-r",
             "Deletes the entry for the domain.") do |value|
        options[:remove] = true
      end
      opt.on("--list", "-l",
             "List all the domains in the databse.") do |value|
        options[:list] = true
      end
      opt.on("--change-password", "-p",
             "Changes the master password.") do |value|
        options[:change_password] = true
      end
      opt.on("--dump", "-d", 
             "If you specify a domain, this causes the ",
             "login info to be printed out directly in ",
             "the console, instead of via 'less'. If you",
             "don't specify a domain, this decrypts and ",
             "dumps the entire password database into a ",
             "YAML string. This string is printed to ",
             "STDOUT, so you can redirect it into a file.",
             "This is to ensure that you can get all your",
             "login data out whenever you want.") do |value|
        options[:dump] = true
      end
    end

    opts.parse! argv
    unless argv.first || options[:dump] || options[:list] || options[:change_password]
      raise OptionParser::InvalidArgument, "You must specify a domain name."
    end
    options
   rescue OptionParser::InvalidArgument, OptionParser::InvalidOption => e
    puts opts
    puts
    puts e
    exit 1
  end

  def self.run(argv = ARGV)
    options = process_args argv
    domain = argv.shift
    @db = db = if database_exists?
      unlock_database
    else
      create_database
    end
    if options[:change_password]
      puts "To change the master password:"
      password = set_password
      @db.change_master_password(password)
      @db.save!
      puts "Master password changed."
    elsif options[:list]
      (domains=db.ui.list).each_with_index do |x,i|
        puts "%2d %s" %[i,x]
      end
      loop do
        print "Type a number to view an entry, or return to exit: "
        reply = STDIN.gets.chomp
        unless reply =~ /\d+/
          puts "Ok, good bye."
          exit
        end
        # lookup the domain
        output = db.ui.decrypt(domains[reply.to_i])
        if output
          IO.popen("less", "w") do |pipe|
            pipe.puts output
          end
        end
      end
    elsif options[:remove]
      puts db.ui.remove(domain)
    elsif options[:dump]
      puts db.ui.dump(domain)
    else
      # lookup the domain
      output = db.ui.decrypt(domain)
      if output
        if options[:edit]
          output = UI::EDIT_HEADER + output
          edit_entry( output )
        else
          IO.popen("less", "w") do |pipe|
            pipe.puts output
          end
        end
      else
        # if no entry, create one
        # We loop so user can add multiple entries without having to enter the
        # password every time.
        loop do
          puts "Creating entry."
          sleep 1
          edit_entry( UI::EDIT_HEADER + db.ui.new_entry(domain) )
          puts "Added entry for #{domain}"
          print "Add another? (y/n) "
          reply = STDIN.gets.chomp
          unless reply =~ /y/i
            puts "Ok, good bye."
            exit
          end
          puts
          print "Domain name for new entry: "
          domain = STDIN.gets.chomp
        end
      end
    end

    exit
    # create database if no database in current dir
    # ask for master password
    # ask for password confirm
    
    if options[:dump]
      # handle dump
    end
    domain = ARGV.shift
    match = decrypt(domain)
    if match
      # show in less
    else
      # create entry
    end
  end

  def self.edit_entry(entry)
    tf = Tempfile.new('simplepass')
    tf.puts( entry )
    tf.close
    system("#{@options[:editor]} #{tf.path}")
    tf.open 
    entry = tf.read 
    @db.ui.add(entry)
    @db.save!
    tf.close(true)
  end

  def self.database_exists?
    File.exist?('simplepass.db')
  end

  def self.unlock_database
    loop do
      print "Master password: "
      `stty -echo`
      password = STDIN.gets.chomp
      `stty echo`
      puts
      exit if password == ''
      db = SimplePass.new(password, 'simplepass.db')
      begin
        db.load
        break db
      rescue Exception
        puts "Sorry, that password isn't correct." 
      end
    end
  end

  def self.create_database
    print "Can't find a simplepass.db database in this directory. Create one? (y/n) "
    reply = STDIN.gets.chomp
    puts

    unless reply =~ /y/i
      puts "Ok, good bye."
      exit
    end
    password = set_password
    puts "Creating database simplepass.db and encrypting it with the master password."
    db = SimplePass.new(password, 'simplepass.db')
    db.save!
    db
  end

  def self.set_password
    password = nil
    loop do
      print "Please type a master password (will not show what you type): "
      `stty -echo`
      password = STDIN.gets.chomp
      puts 
      print "Please confirm by typing it again: "
      password_confirm = STDIN.gets.chomp
      puts
      unless password == password_confirm
        puts "Sorry, you passwords do not match."
      else
        `stty echo`
        break
      end
    end
    password
  end

  HEADER = "This is an encrypted SimplePass password file. Do not edit this file directly.\n"
  DIVIDER = "====BEGIN DATA===\n" 
  attr_accessor :master_password, :header
  def initialize(master_password, file=nil)
    @master_password = master_password
    @blowfish = Crypt::Blowfish.new(@master_password)
    @file = file
    @header = HEADER
    super({})
  end

  def change_master_password(password)
    @master_password = password
    @blowfish = Crypt::Blowfish.new(@master_password)
  end

  def encrypt(plain_string)
    @blowfish.encrypt_string(plain_string)
  end

  def decrypt(encrypted_string)
    @blowfish.decrypt_string(encrypted_string)
  end

  def load
    File.open(@file, "r") do |file|
      content = file.read
      # strip the header
      @header, data = content.split(DIVIDER, 2)
      hash = YAML::load(decrypt(data))
      hash.each_pair {|k,v| self[k] = v}
    end
  end

  def self.load(master_password, file)
    simplepass = self.new(master_password, file)
    simplepass.load
    simplepass
  end

  def save!
    File.open(@file, "w") do |file|
      # Write an informative header
      file.write(@header + DIVIDER)
      # Dump the encrypted data
      dump = YAML::dump(self)
      file.write(encrypt(dump))
    end
  end
  alias_method :save, :save!

  def ui
    @ui ||= UI.new(self)
    @ui
  end

  class UI
    EDIT_HEADER = 
"Fill the the fields below. Leave the first line, with the name of the domain,
untouched. The notes field can accept multiple lines, including blank lines.\n" + ('-' * 20 + "\n")

    def initialize(simplepass)
      @database = simplepass
    end

    def decrypt(domain)
      matching_key = @database.keys.detect {|k| k == domain}
      return nil unless matching_key
      # format the YAML to remove weird symbol markup
      result = @database[matching_key]
      formatted_result = "#{matching_key}\nlogin: #{result[:login]}\npassword: #{result[:password]}\nnotes: #{result[:notes]}\n"
    end

    def list
      @database.keys.sort
    end

    def strip_header(text)
      text.split(('-' * 20) + "\n", 2)[-1]
    end

    def add(text)
      entry = {}
      # strip off header text, which is just for instructions
      text = strip_header(text)
      # split text into lines
      # first line is the domain
      domain = text.split("\n")[0].chomp
      login = text.match(/^login:(.*)/)[1].strip
      password = text.match(/^password:(.*)/)[1].strip
      notes = text.split(/^notes:\s*/, 2)[1].strip
      @database[domain] = {:login => login, :password => password, :notes => notes}
    end
  
    def dump(domain=nil)
      unless domain
        return @database.to_yaml
      end
      @database[domain].to_yaml
    end

    def new_entry(domain)
      return <<END
#{domain}
login:  
password:  
notes:  
END
    end

    def remove(domain)
      unless decrypt(domain)
        return "There is no entry for #{domain} in the database."
      end
      @database.delete(domain)
      @database.save!
      puts = "#{domain} deleted."
    end

  end

end

if __FILE__ == $0
  SimplePass.run ARGV
end
