require File.dirname(__FILE__) + "/../lib/simplepass.rb"
require 'fileutils'

data_dir = File.dirname(__FILE__) + '/data'

describe "SimplePass: Core Engine" do

  before do
    @master_password = "betahouse"
    @test_file = data_dir + "/testSimplePass.txt"
    @simplepass = SimplePass.new(@master_password, @test_file)
  end

  it "should encrypt and decrypt a plain text string" do
    # These are lower level methods
    encrypted = @simplepass.encrypt("hello world")
    decrypted = @simplepass.decrypt(encrypted)
    decrypted.should == "hello world"
  end

  it "should act like a hash of entries, keyed by domain name" do
    # Use these higher level methods
    @simplepass['yahoo.com'] = {:login => "pasta", :password => "pizza"}
    @simplepass['yahoo.com'][:password].should == 'pizza'
    @simplepass['yahoo.com'][:login].should == 'pasta'
    @simplepass['google.com'].should be_nil
  end

  it "should be able to retrieve a password by key (domain)" do
    @simplepass['yahoo.com'] = {:login => "pasta", :password => "pizza"}
    @simplepass['yahoo.com'].should == {:login => "pasta", :password => "pizza"}
  end

  it "should save the passwords into an encypted YAML file" do
    @simplepass['yahoo.com'] = {:login => "pasta", :password => "pizza"}
    @simplepass['flickr.com'] = {:login => "apple", :password => "orange", :note => "hello!"}

    @simplepass.save!

    new_simplepass = SimplePass.load("betahouse", @test_file) 
    new_simplepass['yahoo.com'][:login].should == 'pasta'
    new_simplepass['flickr.com'][:note].should == "hello!"
    new_simplepass['flickr.com'][:password].should == "orange"
  end

  it "should put a default unecrypted header section on the file" do
    @simplepass.header.should == SimplePass::HEADER 
  end

  it "should fail to decrypt a file with the wrong password" do
    @simplepass['yahoo.com'] = {:login => "pasta", :password => "pizza"}
    @simplepass.save!
    lambda do
      @simplepass2 = SimplePass.new("wrong password", @test_file)
      @simplepass2.load
    end.should raise_error(Exception)
    @simplepass3 = SimplePass.new("betahouse", @test_file).load
    @simplepass3['yahoo.com'].should == {:login => "pasta", :password => "pizza"}
  end

  after(:all) do
    FileUtils.rm(@test_file)
  end
end

# Use less as the viewer because it does not leave traces of the displayed 
# content in the bash shell history.
describe "User Interface: a user" do
  before do
    @test_file = data_dir + "/testSimplePass.txt"
    @master_password = "betahouse"
    @simplepass = SimplePass.new(@master_password, @test_file)
    @simplepass['yahoo.com'] = {:login => "pasta", :password => "pizza"}
    @simplepass['google.com'] = {:login => "glue", :password => "cheese"}
    @simplepass['goo.com'] = {:login => "dog", :password => "rock"}
    @simplepass.keys.size.should == 3
    @ui = @simplepass.ui
  end

  it "should be able to retrieve a password as a YAML hash" do
    @ui.decrypt('yahoo.com').should == "yahoo.com\nlogin: pasta\npassword: pizza\nnotes: \n"
  end

  it "should retrieve an empty if no matches" do
    @ui.decrypt('blah').should be_nil
  end

  it "should read in a YAML and add it to the password database" do
    entry = <<-yaml
header text - will be stripped
--------------------
dogster.com
login: u2
password: hello
notes: my dog
is the best
    yaml
    @ui.add(entry)
    @simplepass['dogster.com'].should == {:login => 'u2', :password => 'hello', :notes => "my dog\nis the best"}
    @ui.decrypt('dogster.com').should == "dogster.com\nlogin: u2\npassword: hello\nnotes: my dog\nis the best\n"
    @simplepass.keys.size.should == 4
  end

  it "should decrypt the entire password database" do
    @ui.dump.should == @simplepass.to_yaml
  end

  it "should prepare a form for a new entry" do
    @ui.new_entry('amazon.com').should == "amazon.com\nlogin:  \npassword:  \nnotes:  \n"
  end

  it "should delete a domain" do
    @ui.remove("goo.com").should == "goo.com deleted."
    @simplepass['goo.com'].should be_nil
  end

  after(:all) do
    FileUtils.rm(@test_file) 
  end
end

